package org.ow2.bonita.example.aw;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.security.auth.login.LoginContext;
import javax.security.auth.login.LoginException;

import org.ow2.bonita.facade.ManagementAPI;
import org.ow2.bonita.facade.QueryRuntimeAPI;
import org.ow2.bonita.facade.RuntimeAPI;
import org.ow2.bonita.facade.def.majorElement.ProcessDefinition;
import org.ow2.bonita.facade.runtime.ActivityInstance;
import org.ow2.bonita.facade.runtime.ActivityState;
import org.ow2.bonita.facade.runtime.ProcessInstance;
import org.ow2.bonita.facade.runtime.TaskInstance;
import org.ow2.bonita.facade.uuid.PackageDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessInstanceUUID;
import org.ow2.bonita.facade.uuid.TaskUUID;
import org.ow2.bonita.util.AccessorUtil;
import org.ow2.bonita.util.BonitaConstants;
import org.ow2.bonita.util.BonitaException;
import org.ow2.bonita.util.BonitaRuntimeException;
import org.ow2.bonita.util.Misc;
import org.ow2.bonita.util.SimpleCallbackHandler;
import org.ow2.bonita.util.StandardCallbackHandler;

/**
 * 
 * @author Charles SOUILLARD
 *
 */
public final class MultiInstantiationApproval {

  private MultiInstantiationApproval() { }

  private static final String BAR_PREFIX = "-bar=";
  private static final String NB_PREFIX = "-nb=";
  
  private static final Logger LOG = Logger.getLogger(MultiInstantiationApproval.class.getName());
  public static final String PROCESS_ID = "multiInstantiation";

  private static Map<String, String> getVotes(int numberOfAccept) {
    Map<String, String> votes = new HashMap<String, String>();
    switch (numberOfAccept) {
      case 0:
        votes.put("john", "no");
        votes.put("jack", "no");
        votes.put("james", "no");
        break;
      case 1 :
        votes.put("john", "yes");
        votes.put("jack", "no");
        votes.put("james", "no");
        break;
      case 2 :
        votes.put("john", "yes");
        votes.put("jack", "yes");
        votes.put("james", "no");
        break;
      default:
      case 3 :
        votes.put("john", "yes");
        votes.put("jack", "yes");
        votes.put("james", "yes");
        break;
    }
    return votes;
  }

  private static void performTask(ProcessInstanceUUID instanceUUID, String vote) throws BonitaException {
    final QueryRuntimeAPI queryRuntimeAPI = AccessorUtil.getQueryRuntimeAPI();
    final RuntimeAPI runtimeAPI = AccessorUtil.getRuntimeAPI();

    final Collection<ActivityInstance< TaskInstance >> activities = 
      queryRuntimeAPI.getTaskList(instanceUUID, ActivityState.READY);

    if (activities.isEmpty()) {
      throw new BonitaRuntimeException("No task found? Bad User?");
    }
    for (ActivityInstance< TaskInstance > activity : activities) {
      final TaskUUID taskUUID = activity.getBody().getUUID();
      final String activityId = activity.getActivityId();
      log("Starting task associated to activity: " + activityId);
      runtimeAPI.startTask(taskUUID, true);
      runtimeAPI.setActivityInstanceVariable(activity.getUUID(), "decisionAccepted", vote);
      log("Finishing task associated to activity: " + activityId + ": vote = " + vote);
      runtimeAPI.finishTask(taskUUID, true);
      log("Task associated to activity: " + activityId + " finished.");
    }
  }

  public static ProcessInstanceUUID execute(final byte[] barFile, final int numberOfAccept, 
      final Map<String, String> logins, LoginContext loginContext) throws BonitaException {
    //get APIs
    final RuntimeAPI runtimeAPI = AccessorUtil.getAPIAccessor().getRuntimeAPI();
    final ManagementAPI managementAPI = AccessorUtil.getAPIAccessor().getManagementAPI();

    //deployment
    final ProcessDefinition process = managementAPI.deployBar(barFile).get(PROCESS_ID);
    final PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    Misc.badStateIfNull(packageUUID, "packageUUID is null in process : " + process.getProcessId());

    //instantiation
    final ProcessInstanceUUID instanceUUID = runtimeAPI.instantiateProcess(process.getUUID());

    Map<String, String> votes = getVotes(numberOfAccept);

    try {
      while (!votes.isEmpty()) {
        loginContext.logout();
        String user = votes.keySet().iterator().next();
        String currentVote;
        if (logins.containsKey(user)) {
          loginContext = new LoginContext("Bonita", new SimpleCallbackHandler(user, logins.get(user)));
          loginContext.login();
          currentVote = votes.get(user);
        } else {
          String msg = "Please login as a separate user to start task. Tasks available for users: ";
          for (String v : votes.keySet()) {
            msg += v + " ";
          }
          System.out.println(msg);
          loginContext.login();
          System.out.println("Will you accept the request ? [yes/no]");
          try {
            currentVote = new BufferedReader(new InputStreamReader(System.in)).readLine();
            if (!"yes".equals(currentVote) && !"no".equals(currentVote)) {
              throw new BonitaRuntimeException("invalid value " + currentVote);
            }
          } catch (IOException e) {
            throw new BonitaRuntimeException(e);
          }
        }
        ///// Find current user
        // TODO: find a better solution
        Collection<ActivityInstance<TaskInstance>> userTasks = 
          AccessorUtil.getAPIAccessor().getQueryRuntimeAPI().getTaskList(ActivityState.READY);
        if (userTasks == null || userTasks.isEmpty()) {
          throw new BonitaRuntimeException("No task found for current user.");
        }
        String currentUser = userTasks.iterator().next().getBody().getTaskUser();
        votes.remove(currentUser);
        //// End of find current user
        performTask(instanceUUID, currentVote);
      }
    } catch (LoginException e) {
      throw new BonitaRuntimeException(e);
    }
    return instanceUUID;
  }

  private static void usage(final String[] args) {
    LOG.severe("Usage: java " + MultiInstantiationApproval.class.getName() 
        + " " + BAR_PREFIX + "barFile " + NB_PREFIX + "numberOfAccept");
    LOG.severe("");
    LOG.severe("Actual args:");
    if (args == null) {
      LOG.severe("  args is null");  
    } else {
      for (String arg : args) {
        LOG.severe("  arg:" + arg);
      }
    }
    throw new BonitaRuntimeException("Please read the above usage");
  }
  
  public static void main(final String[] args) throws Exception {
    if (args == null || args.length != 2) {
      usage(args);
    }
    
    byte[] barFile = null;
    int numberOfAccept = 0;
    for (String arg : args) {
      if (arg.startsWith(BAR_PREFIX)) {
        barFile = Misc.getAllContentFrom(new File(arg.substring(BAR_PREFIX.length())));
      } else if (arg.startsWith(NB_PREFIX)) {
        numberOfAccept = Integer.parseInt(arg.substring(NB_PREFIX.length()));
      } else {
        usage(args);
      }
    }

    final String loginMode = System.getProperty(BonitaConstants.LOGIN_MODE_PROPERTY);
    LoginContext loginContext = null;
    if (loginMode != null && BonitaConstants.LOGIN_MODE_TEST.equals(loginMode)) {
      loginContext = new LoginContext("Bonita", new SimpleCallbackHandler("john", "bpm"));
    } else {
      loginContext = new LoginContext("Bonita", new StandardCallbackHandler());
    }
    loginContext.login();
    
    ProcessInstanceUUID instanceUUID = execute(barFile, numberOfAccept, new HashMap<String, String>(), loginContext);

    cleanPackage(instanceUUID);

    loginContext.logout();

    if (LOG.isLoggable(Level.INFO)) {
      LOG.info("\n***** End of ApprovalWorkflow Main. *****\n");
    }
  }


  public static void cleanPackage(ProcessInstanceUUID instanceUUID) throws BonitaException {
    final ManagementAPI managementAPI = AccessorUtil.getManagementAPI();
    final QueryRuntimeAPI queryRuntimeAPI = AccessorUtil.getQueryRuntimeAPI();
    final ProcessInstance instance = queryRuntimeAPI.getProcessInstance(instanceUUID);
    final PackageDefinitionUUID packageUUID = instance.getPackageDefinitionUUID();

    //undeployment
    managementAPI.undeploy(packageUUID);
    //journal + history cleaning
    managementAPI.deletePackage(packageUUID);
  }

  private static void log(String msg) {
    if (LOG.isLoggable(Level.INFO)) {
      LOG.info(Misc.LINE_SEPARATOR + "***** " + msg + " *****" + Misc.LINE_SEPARATOR);
    }
  }
  
}
